import time
import warnings

import napari
import skimage.draw
from napari.utils.events import Event
from qtpy.QtWidgets import QSpinBox, QVBoxLayout, QCheckBox, QLabel, QComboBox
from qtpy.QtCore import QMutex, QThread, QObject, Signal

from ._utils.progress_widget import ProgressWidget
from ._utils.widget_with_layer_list import WidgetWithLayerList
from .image_processing_widget import ImageProcessingWidget
from ..minimal_contour import MinimalContourCalculator
import numpy as np
from napari.layers import Points, Image, Labels
from napari.layers.points._points_constants import Mode
from skimage.filters import gaussian

GRADIENT_BASED = 0
INTENSITY_BASED = 2


def bbox_around_points(pts):
    p1 = pts.min(0)
    p2 = pts.max(0)
    size = p2 - p1
    from_i = p1 - size[0]*0.1 - 10
    to_i = p2 + size[0]*0.1 + 10
    return from_i, to_i


class MinimalContourWidget(WidgetWithLayerList):
    def __init__(self, viewer: napari.Viewer):
        super().__init__(viewer, [("image", Image), ("labels", Labels), ("anchor_points", Points)])
        self.viewer = viewer
        self.calculator = MinimalContourCalculator(None, 3)
        self.progress_dialog = ProgressWidget(message="Drawing mask...")
        self.move_mutex = QMutex()
        self.draw_worker = self.DrawWorker()
        self.draw_thread = QThread()
        self.draw_worker.moveToThread(self.draw_thread)
        self.draw_worker.done.connect(self.draw_thread.quit)
        self.draw_worker.done.connect(self.set_mask)
        self.draw_worker.done.connect(lambda: self.progress_dialog.setVisible(False))
        self.draw_thread.started.connect(self.draw_worker.run)
        self._img = None
        self.point_triangle = np.zeros((3, 2), dtype=np.float64) - 1  # start point, current position, end point
        self.remove_last_anchor = False
        self.last_added_with_shift = None
        self.last_segment_length = None
        self.anchor_points.combobox.currentIndexChanged.connect(self.set_callbacks)
        self.image.combobox.currentIndexChanged.connect(self.set_image)
        self.feature_inverted = False
        self.test_script = False
        layout = QVBoxLayout()

        layout.addWidget(QLabel("Used feature"))
        self.feature_dropdown = QComboBox()
        self.feature_dropdown.addItems(["High gradient", "High intensity", "Low intensity", "Custom"])
        self.feature_dropdown.currentIndexChanged.connect(self.on_feature_change)
        layout.addWidget(self.feature_dropdown)

        self.feature_editor = ImageProcessingWidget(self._img, viewer)
        self.feature_editor.setVisible(self.feature_dropdown.currentText() == "Custom")
        self.feature_editor.script_worker.done.connect(self.set_features)
        def set_test_script():
            self.test_script = True
        self.feature_editor.try_script.connect(set_test_script)
        layout.addWidget(self.feature_editor)

        layout.addWidget(QLabel("Param"))
        self.param_spinbox = QSpinBox()
        self.param_spinbox.setMinimum(1)
        self.param_spinbox.setMaximum(50)
        self.param_spinbox.setValue(5)
        self.param_spinbox.valueChanged.connect(lambda val: self.calculator.set_param(val))
        self.calculator.set_param(self.param_spinbox.value())
        layout.addWidget(self.param_spinbox)

        layout.addWidget(QLabel("Increment label id"))
        self.autoincrease_label_id_checkbox = QCheckBox()
        self.autoincrease_label_id_checkbox.setChecked(True)
        layout.addWidget(self.autoincrease_label_id_checkbox)

        layout.addWidget(QLabel("Point size"))
        self.point_size_spinbox = QSpinBox()
        self.point_size_spinbox.setMinimum(1)
        self.point_size_spinbox.setMaximum(50)
        self.point_size_spinbox.setValue(2)
        layout.addWidget(self.point_size_spinbox)
        layout.addStretch()
        self.setLayout(layout)

        self.set_filter_func()

        viewer.layers.events.connect(self.invalidate_filter)

        def change_layer_callback(num):
            def change_layer(_):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    if viewer.window.qt_viewer.layers.model().hasIndex(num - 1, 0):
                        index = viewer.window.qt_viewer.layers.model().index(num - 1, 0)
                        layer_name = viewer.window.qt_viewer.layers.model().data(index)
                        viewer.layers.selection.select_only(viewer.layers[layer_name])
            return change_layer
        for i in range(1, 10):
            viewer.bind_key("Control-%d" % i, overwrite=True)(change_layer_callback(i))
        self.from_e_points_layer = viewer.add_points(
            ndim=2,
            name="from E [DO NOT TOUCH] <hidden>",
            size=self.point_size_spinbox.value(),
            face_color="red",
            edge_color="red"
        )
        self.to_s_points_layer = viewer.add_points(
            ndim=2,
            name="to S [DO NOT TOUCH] <hidden>",
            size=self.point_size_spinbox.value(),
            face_color="gray",
            edge_color="gray"
        )
        self.output = viewer.add_points(
            ndim=2,
            name="temp [DO NOT TOUCH] <hidden>",
            size=self.point_size_spinbox.value()
        )
        self.shift_down = False
        self.ctrl_down = False
        self.point_size_spinbox.valueChanged.connect(self.change_point_size)
        self.change_point_size(self.point_size_spinbox.value())
        viewer.dims.events.current_step.connect(self.set_image)

    def set_features(self, data):
        if self.test_script:
            self.test_script = False
            return
        if data is None:
            return
        self.calculator.set_image(data)

    def showEvent(self, e):
        super().showEvent(e)
        if "Anchors [DO NOT TOUCH]" in self.viewer.layers:
            return
        anchors = self.viewer.add_points(ndim=2, name="Anchors [DO NOT TOUCH]", symbol="x")
        self.anchor_points.layer = anchors
        self.anchor_points.combobox.setEnabled(False)
        self.viewer.layers.events.connect(self.move_temp_to_top)

    def invalidate_filter(self, e):
        # if e.type in ["highlight", "mode", "set_data", "data", "thumbnail", "loaded"]:
        #     return
        self.set_filter_func()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.viewer.window.qt_viewer.layers.model().invalidateFilter()
            # self.set_filter_func()

    def set_filter_func(self):
        def filter_func(row, parent):
            return row < len(self.viewer.layers) and "<hidden>" not in self.viewer.layers[row].name
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.viewer.window.qt_viewer.layers.model().filterAcceptsRow = filter_func

    def change_point_size(self, size):
        self.to_s_points_layer.size = size
        self.to_s_points_layer.selected_data = {}
        self.to_s_points_layer.current_size = size

        self.from_e_points_layer.size = size
        self.from_e_points_layer.selected_data = {}
        self.from_e_points_layer.current_size = size

        self.output.size = size
        self.output.selected_data = {}
        self.output.current_size = size
        self.viewer.window.qt_viewer.canvas.native.setFocus()

    def shift_pressed(self, _):
        self.shift_down = True
        self.from_e_points_layer.face_color = "gray"
        self.from_e_points_layer.current_face_color = "gray"
        self.to_s_points_layer.face_color = "red"
        self.to_s_points_layer.current_face_color = "red"
        self.from_e_points_layer.edge_color = "gray"
        self.from_e_points_layer.current_edge_color = "gray"
        self.to_s_points_layer.edge_color = "red"
        self.to_s_points_layer.current_edge_color = "red"
        yield
        self.shift_down = False
        self.from_e_points_layer.face_color = "red"
        self.from_e_points_layer.current_face_color = "red"
        self.to_s_points_layer.face_color = "gray"
        self.to_s_points_layer.current_face_color = "gray"
        self.from_e_points_layer.edge_color = "red"
        self.from_e_points_layer.current_edge_color = "red"
        self.to_s_points_layer.edge_color = "gray"
        self.to_s_points_layer.current_edge_color = "gray"

    def ctrl_pressed(self, _):
        self.ctrl_down = True
        yield
        self.ctrl_down = False

    def ctrl_shift_pressed(self, _):
        shift_callback = self.shift_pressed(_)
        yield from shift_callback
        yield from shift_callback

    def shift_ctrl_pressed(self, _):
        ctrl_callback = self.ctrl_pressed(_)
        yield from ctrl_callback
        yield from ctrl_callback

    def esc_callback(self, _):
        self.clear_all()

    def clear_all(self):
        self.output.data = np.empty((0, 2), dtype=self.output.data.dtype)
        self.anchor_points.layer.data = np.empty((0, 2), dtype=self.output.data.dtype)
        self.from_e_points_layer.data = np.empty((0, 2), dtype=self.output.data.dtype)
        self.to_s_points_layer.data = np.empty((0, 2), dtype=self.output.data.dtype)
        self.point_triangle[:] = -1

    def move_temp_to_top(self, e):
        if e.type in ["highlight", "mode", "set_data", "data", "thumbnail", "loaded"]:
            return
        layer_list = e.source
        with layer_list.events.moved.blocker(), layer_list.events.moving.blocker():
            temp_idx = layer_list.index(self.output)
            if temp_idx != len(layer_list) - 1:
                layer_list.move(temp_idx, -1)
            if self.anchor_points.layer is not None:
                points_idx = layer_list.index(self.anchor_points.layer)
                if points_idx != len(layer_list) - 2:
                    layer_list.move(points_idx, -2)
            to_s_idx = layer_list.index(self.to_s_points_layer)
            if to_s_idx != len(layer_list) - 3:
                layer_list.move(to_s_idx, -3)
            from_e_idx = layer_list.index(self.from_e_points_layer)
            if from_e_idx != len(layer_list) - 4:
                layer_list.move(from_e_idx, -4)

    def estimate(self, image: np.ndarray):
        from_i, to_i = bbox_around_points(self.point_triangle)
        from_i = np.clip(from_i, 0, np.asarray(image.shape[:2])).astype(int)
        to_i = np.clip(to_i, 0, np.asarray(image.shape[:2])).astype(int)
        self.calculator.set_boundaries(from_i[1], from_i[0], to_i[1], to_i[0])
        results = self.calculator.run(
            self.point_triangle,
            True,
            True,
            True
        )
        return results

    def on_feature_change(self, _):
        current_text = self.feature_dropdown.currentText()
        self.feature_editor.setVisible(current_text == "Custom")
        if current_text.startswith("Low") != self.feature_inverted:
            self.feature_inverted = current_text.startswith("Low")
        self.set_image()
        self.calculator.set_method(GRADIENT_BASED if "gradient" in current_text else INTENSITY_BASED)

    def on_mouse_move(self, layer, event):
        if not self.move_mutex.tryLock():
            return
        try:
            if self.image.layer is None or self.anchor_points.layer is None:
                return
            if layer.mode != Mode.ADD:
                return
            self.point_triangle[1] = [event.position[i] for i in range(len(event.position)) if i in event.dims_displayed]
            if np.any(self.point_triangle < 0) or np.any(self.point_triangle >= self._img.shape[:2]):
                return
            if not self.ctrl_down:
                if self.feature_editor.isVisible():
                    if self.feature_editor.features is None:
                        warnings.warn("Feature image not calculated. Run script using the 'Set' button.")
                        return
                    results = self.estimate(self.feature_editor.features)
                else:
                    if self._img is None:
                        warnings.warn("Image was not set.")
                        return
                    results = self.estimate(self._img)
            else:
                results = [
                    np.vstack(skimage.draw.line(
                        int(self.point_triangle[1, 0]),
                        int(self.point_triangle[1, 1]),
                        int(self.point_triangle[0, 0]),
                        int(self.point_triangle[0, 1])

                    )).T,
                    np.vstack(skimage.draw.line(
                        int(self.point_triangle[2, 0]),
                        int(self.point_triangle[2, 1]),
                        int(self.point_triangle[1, 0]),
                        int(self.point_triangle[1, 1])
                    )).T
                ]
            self.from_e_points_layer.data = np.flipud(results[0])
            self.from_e_points_layer.selected_data = {}
            self.to_s_points_layer.data = np.flipud(results[1])
            self.to_s_points_layer.selected_data = {}
        finally:
            self.move_mutex.unlock()

    def set_callbacks(self):
        if self.anchor_points.layer is None:
            return
        self.anchor_points.layer.events.data.connect(self.data_event)
        self.anchor_points.layer.bind_key("Shift", overwrite=True)(self.shift_pressed)
        self.anchor_points.layer.bind_key("Control", overwrite=True)(self.ctrl_pressed)
        self.anchor_points.layer.bind_key("Control-Shift", overwrite=True)(self.ctrl_shift_pressed)
        self.anchor_points.layer.bind_key("Shift-Control", overwrite=True)(self.shift_ctrl_pressed)
        self.anchor_points.layer.bind_key("Escape", overwrite=True)(self.esc_callback)
        if self.on_double_click not in self.anchor_points.layer.mouse_double_click_callbacks:
            self.anchor_points.layer.mouse_double_click_callbacks.append(self.on_double_click)
        if self.on_mouse_move not in self.anchor_points.layer.mouse_move_callbacks:
            self.anchor_points.layer.mouse_move_callbacks.append(self.on_mouse_move)
        if self.on_right_click not in self.anchor_points.layer.mouse_drag_callbacks:
            self.anchor_points.layer.mouse_drag_callbacks.insert(0, self.on_right_click)

    def set_image(self):
        image_layer: Image = self.image.layer
        if image_layer is None:
            self._img = None
            self.feature_editor.image = None
            self.feature_editor.features = None
            return
        if not image_layer.visible:
            image_layer.set_view_slice()
        self.autoincrease_label_id_checkbox.setChecked(image_layer.ndim == 2)
        image = image_layer._data_view.astype(float)
        image = gaussian(image, channel_axis=2 if image.ndim == 3 else None)
        if image.ndim == 2:
            image = np.concatenate([image[..., np.newaxis]] * 3, axis=2)
        image = (image - image.min()) / (image.max() - image.min())
        if self.feature_inverted:
            image = 1 - image
        if self.feature_editor.isVisible():
            self.feature_editor.image = image
        else:
            self.calculator.set_image(image)
        self._img = image

    def data_event(self, event):
        if event.source != self.anchor_points.layer or len(self.anchor_points.layer.data) == 0:
            return
        if self.remove_last_anchor:
            with self.anchor_points.layer.events.data.blocker():
                self.anchor_points.layer.data = self.anchor_points.layer.data[:-1]
                self.remove_last_anchor = False
                return
        anchor_data = self.anchor_points.layer.data
        anchor_data[-1] = np.clip(anchor_data[-1], 0, self._img.shape[:2])
        if len(anchor_data) > 1 and np.all(np.round(anchor_data[-1]) == np.round(anchor_data[-2])):
            with self.anchor_points.layer.events.data.blocker():
                self.anchor_points.layer.data = self.anchor_points.layer.data[:-1]
            self.anchor_points.layer.refresh()
            return
        self.anchor_points.layer.refresh()
        if self.shift_down:
            self.last_added_with_shift = True
            with self.anchor_points.layer.events.data.blocker():
                self.anchor_points.layer.data = np.roll(self.anchor_points.layer.data, 1, 0)
            if len(self.to_s_points_layer.data):
                self.output.data = np.concatenate([self.to_s_points_layer.data, self.output.data], 0)
                self.last_segment_length = len(self.to_s_points_layer.data)
        else:
            if len(self.from_e_points_layer.data):
                self.output.data = np.concatenate([self.output.data, self.from_e_points_layer.data], 0)
                self.last_added_with_shift = False
                self.last_segment_length = len(self.from_e_points_layer.data)

        self.point_triangle[-1] = self.anchor_points.layer.data[0]
        self.point_triangle[0] = self.anchor_points.layer.data[-1]

    def on_double_click(self, *args):
        if self.shift_down and len(self.from_e_points_layer.data):
            self.output.data = np.concatenate([self.output.data, self.from_e_points_layer.data], 0)
        elif not self.shift_down and len(self.to_s_points_layer.data):
            self.output.data = np.concatenate([self.to_s_points_layer.data, self.output.data], 0)
        self.points_to_mask()



    def on_right_click(self, layer, event: Event):
        if event.button == 2 and layer.mode == Mode.ADD:
            self.remove_last_anchor = True
            if self.last_segment_length is not None:
                with self.anchor_points.layer.events.data.blocker():
                    if self.last_added_with_shift:
                        self.anchor_points.layer.data = self.anchor_points.layer.data[1:]
                        self.output.data = self.output.data[self.last_segment_length:]
                    else:
                        self.anchor_points.layer.data = self.anchor_points.layer.data[:-1]
                        self.output.data = self.output.data[:-self.last_segment_length]
                self.point_triangle[-1] = self.anchor_points.layer.data[0]
                self.point_triangle[0] = self.anchor_points.layer.data[-1]
                self.last_segment_length = None

    class DrawWorker(QObject):
        done = Signal("PyQt_PyObject")
        contour: np.ndarray
        mask_shape: tuple


        def run(self):
            mask = skimage.draw.polygon2mask(self.mask_shape, self.contour)
            self.done.emit(mask)

    def set_mask(self, mask):
        self.labels.layer._slice.image.raw[mask] = self.labels.layer.selected_label
        self.labels.layer.events.data()
        self.labels.layer.refresh()
        self.clear_all()
        if self.labels.layer and self.autoincrease_label_id_checkbox.isChecked():
            self.labels.layer.selected_label += 1

    def points_to_mask(self):
        if self._img is None or len(self.output.data) == 0:
            return
        elif self.labels.layer is None:
            warnings.warn("Missing output labels layer.")
            return
        if not self.labels.layer.visible:
            self.labels.layer.set_view_slice()
        self.progress_dialog.setVisible(True)
        self.draw_worker.contour = self.output.data
        self.draw_worker.mask_shape = self._img.shape[:2]
        self.draw_thread.start()



import os.path
import warnings

import cv2
import napari
import numpy as np

from qtpy.QtCore import QEvent, Qt, QObject
from qtpy.QtGui import QImage, QPixmap, QHideEvent, QShowEvent
from qtpy.QtWidgets import QWidget, QVBoxLayout, QLineEdit, QHBoxLayout, QLabel, QListWidgetItem, QListWidget, QMenu, \
    QComboBox, QPushButton, QStackedWidget, QDockWidget, QFileDialog, QSpinBox
from napari import Viewer
from napari.layers import Image, Labels
from napari.layers.labels._labels_constants import Mode
from scipy.ndimage import find_objects

from ._utils.widget_with_layer_list import WidgetWithLayerList
from .projections import SliceDisplayWidget
from ..boundingbox import BoundingBoxLayer

import itertools
import csv

import PIL.Image as PilImage

import time


def slice_len(slice_, count=None):
    if count is None:
        count = slice_.stop
    return len(range(*slice_.indices(count)))


class QObjectWidget(QWidget):
    def __init__(self, icon=None, name="Object", index=0, bounding_box=None, parent=None):
        super().__init__()
        self.textQVBoxLayout = QVBoxLayout()
        self.textUpQLabel = QLineEdit()
        self.textUpQLabel.setToolTip("Object name")
        self.id_label = QLabel()
        self.id_label.setText("Id: " + str(index))
        self.id_label.setToolTip("label index")
        self.textDownQLabel = QLabel()
        self.textQVBoxLayout.addWidget(self.textUpQLabel)
        self.textQVBoxLayout.addWidget(self.id_label)
        self.allQHBoxLayout  = QHBoxLayout()
        self.iconQLabel      = QLabel()
        self.allQHBoxLayout.addWidget(self.iconQLabel, 0)
        self.allQHBoxLayout.addLayout(self.textQVBoxLayout, 1)
        self.setLayout(self.allQHBoxLayout)
        if icon is not None:
            self.iconQLabel.setPixmap(icon)
        self.setTextUp(name)
        if bounding_box is not None:
            self.setTextDown(bounding_box)
        self.name = name
        self.parent = parent

    @property
    def name(self):
        return self.textUpQLabel.text()

    @name.setter
    def name(self, name):
        self.textUpQLabel.setText(name)

    def setTextUp(self, text):
        self.textUpQLabel.setText(text)

    def setTextDown(self, bb):
        self.textDownQLabel.setText(("corner 1: %s\ncorner 2: %s" % tuple(bb)))

    def setIcon(self, icon):
        self.iconQLabel.setPixmap(icon)
        self.iconQLabel.setFixedSize(icon.size())


class QObjectListWidgetItem(QListWidgetItem):
    def __init__(self, name, bounding_box, index, parent, viewer, image_layer, mask_layer, channels_dim, *args):
        super().__init__(None, *args)
        if name is None:
            name = parent.name_template
        self.object_item = QObjectWidget(name=name, index=index, bounding_box=bounding_box, parent=self)
        self.setSizeHint(self.object_item.sizeHint())
        self.setFlags(self.flags() | Qt.ItemFlag.ItemIsEditable)
        self.parent = parent
        self.viewer = viewer
        self.image_layer = image_layer
        self.mask_layer = mask_layer
        self.bounding_box = bounding_box
        self.channels_dim = channels_dim

    def __hash__(self):
        return id(self)

    @property
    def idx(self):
        return int(self.object_item.id_label.text().split(":")[-1])

    @idx.setter
    def idx(self, idx):
        self.object_item.id_label.setText("Id: " + str(idx))

    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, icon):
        if len(icon) == 0:
            return
        self._icon = icon.reshape(icon.shape[0], icon.shape[1], -1)
        if self._icon.shape[-1] == 1:
            self._icon = np.tile(self._icon, (1, 1, 3))
        elif self._icon.shape[-1] == 4:
            self._icon = self._icon[..., :-1]
        scale = 64 / max(self._icon.shape[1], self._icon.shape[0])
        out_size = (int(self._icon.shape[1]*scale), int(self._icon.shape[0]*scale))
        self._resized_icon = cv2.resize(self._icon, out_size, interpolation=cv2.INTER_LANCZOS4)
        img = QImage(self._resized_icon, self._resized_icon.shape[1], self._resized_icon.shape[0], self._resized_icon.shape[1] * 3, QImage.Format.Format_RGB888)
        self.pixmap = QPixmap()
        self.pixmap = self.pixmap.fromImage(img, Qt.ImageConversionFlag.ColorOnly)
        self.object_item.setIcon(self.pixmap)
        # self.update_icon()

    @property
    def bounding_box(self):
        return self._bounding_box

    @bounding_box.setter
    def bounding_box(self, bounding_box):
        self._bounding_box = bounding_box
        self.object_item.setTextDown(bounding_box)
        self.object_item.update()
        if self.parent.crop_image_layer is not None and \
            self.parent.selected_idx == self.parent.indexFromItem(self).row():
            self.parent.crop_image_layer.data = self.image_layer.data[self.bbox_idx]
            self.parent.crop_image_layer.translate = tuple(self.bounding_box[0, :])
            self.parent.crop_image_layer.refresh()
            self.parent.crop_mask_layer.data = self.mask_layer.data[self.bbox_idx]
            self.parent.crop_mask_layer.translate = tuple(self.bounding_box[0, :])
            self.parent.crop_mask_layer.refresh()
        self.update_icon()

    @property
    def bbox_idx(self):
        bbox_idx = tuple(
            slice(self.bounding_box[0, d], self.bounding_box[1, d]) for d in range(self.bounding_box.shape[1]))
        return bbox_idx

    @property
    def name(self):
        return self.object_item.name

    @name.setter
    def name(self, name):
        self.object_item.name = name

    def create_layers(self):
        if self.image_layer is not None:
            image_layer = Image(self.image_layer.data[self.bbox_idx],
                                name="Cropped %s %d" % (self.image_layer.name, self.idx),
                                colormap=self.image_layer.colormap, rgb=self.image_layer.rgb)
            image_layer.translate = self.bounding_box[0, :]+self.image_layer.translate
            self.viewer.add_layer(image_layer)
        else:
            image_layer = None
        if self.mask_layer is not None:
            mask_layer = Labels(self.mask_layer.data[self.bbox_idx],
                                name="Cropped %s %d" % (self.mask_layer.name, self.idx))
            mask_layer.selected_label = self.idx
            mask_layer.translate = self.bounding_box[0, :]+self.mask_layer.translate
            mask_layer.brush_size = 1
            mask_layer.mode = Mode.PAINT
            self.mask_layer.events.data.connect(mask_layer.refresh)
            self.viewer.add_layer(mask_layer)
        else:
            mask_layer = None
        return image_layer, mask_layer

    def unselect(self):
        if self.parent.crop_image_layer is not None:
            if self.parent.crop_image_layer in self.viewer.layers:
                self.viewer.layers.remove(self.parent.crop_image_layer)
            self.parent.crop_image_layer = None
        if self.parent.crop_mask_layer is not None:
            if self.parent.crop_mask_layer in self.viewer.layers:
                self.viewer.layers.remove(self.parent.crop_mask_layer)
            if self.mask_layer is not None:
                self.mask_layer.events.data.disconnect(self.parent.crop_mask_layer.refresh)
            self.parent.crop_mask_layer.mouse_drag_callbacks.remove(QObjectListWidgetItem.update_layer)
            self.parent.crop_mask_layer = None
        if self.parent.projections_widget is not None:
            self.viewer.window.remove_dock_widget(self.parent.projections_widget)
            self.parent.projections_widget = None
        if self.mask_layer is not None:
            self.mask_layer.visible = True
        if self.image_layer is not None:
            self.image_layer.visible = True

    def on_select(self):
        self.unselect()
        self.parent.crop_image_layer, self.parent.crop_mask_layer = self.create_layers()
        self.parent.crop_image_layer.name = self.parent.crop_image_layer.name
        self.parent.crop_mask_layer.name = self.parent.crop_mask_layer.name
        self.parent.crop_mask_layer.events.set_data.connect(self.on_data_change)
        if self.image_layer.ndim > 2:
            self.parent.projections_widget = SliceDisplayWidget(self.viewer, self.parent.crop_image_layer, self.parent.crop_mask_layer, self.channels_dim)
            self.viewer.window.add_dock_widget(self.parent.projections_widget)
        self.parent.crop_mask_layer.mouse_drag_callbacks.append(self.update_layer)
        self.viewer.layers.selection.select_only(self.parent.crop_mask_layer)
        self.mask_layer.selected_label = self.idx
        self.mask_layer.visible = False
        self.image_layer.visible = False

    def on_data_change(self, event):
        if self.channels_dim is not None and not self.image_layer.rgb:
            if self.parent.crop_mask_layer._mode == Mode.PAINT:
                data = self.parent.crop_mask_layer.data.max(axis=self.channels_dim, keepdims=True)
                self.parent.crop_mask_layer.data[:] = data
            elif self.parent.crop_mask_layer._mode == Mode.ERASE:
                data = self.parent.crop_mask_layer.data.min(axis=self.channels_dim, keepdims=True)
                self.parent.crop_mask_layer.data[:] = data
        self.mask_layer.data[self.bbox_idx] = self.parent.crop_mask_layer.data
        if self.parent.projections_widget is not None:
            self.parent.projections_widget.update_slider_ranges()
        self.parent.on_layer_event(event)

    @staticmethod
    def update_layer(layer, event):
        yield
        while event.type == "mouse_move":
            yield
        layer.refresh()

    def update_icon(self):
        visible_dims = list(self.viewer.dims.displayed)[-2:]
        bbox_idx = tuple(
            slice(max(self.bounding_box[0, d], 0), self.bounding_box[1, d]) if d in visible_dims
            else self.bounding_box[:, d].mean().astype(int)
            for d in range(self.bounding_box.shape[1])
        )
        if self.image_layer is None:
            icon = np.zeros([slice_len(bbox_idx[d]) for d in range(len(bbox_idx)) if type(bbox_idx[d]) is slice] + [3])
        else:
            icon = self.image_layer.data[bbox_idx]
        if self.image_layer is not None and not self.image_layer.rgb:
            max_ = self.image_layer.contrast_limits[1]
            min_ = self.image_layer.contrast_limits[0]
            icon = np.clip((icon.astype(float)-min_)/(max_ - min_), 0, 1)
            icon = (self.image_layer.colormap.map(icon.ravel()).reshape(icon.shape + (4,)) * 255).astype(np.uint8)[..., :3]
        else:
            icon = icon.astype(np.uint8)[..., :3]
        if self.mask_layer is not None:
            overlay = self.mask_layer.data[bbox_idx] == self.idx
            img = PilImage.fromarray(icon).convert(mode="RGBA")
            mask = PilImage.fromarray(overlay).convert(mode="RGBA")
            mask.putalpha(PilImage.fromarray((overlay*127).astype(np.uint8)))
            img.alpha_composite(mask)
            icon = np.asarray(img)
        self.icon = icon


class ObjectListWidget(QListWidget):
    def __init__(self, viewer, name_template, bounding_box_layer, image_layer, mask_layer, channels_dim, indices=None):
        super().__init__()
        self._bounding_box_layer = None
        self._image_layer = None
        self._mask_layer = None
        self.crop_image_layer = None
        self.crop_mask_layer = None
        self.projections_widget = None
        self._mouse_down = False
        self.selected_idx = None
        self.previous_edge_color = None
        self._indices = None
        self.indices = indices
        self.viewer = viewer
        self.name_template = name_template
        self.channels_dim = channels_dim
        self.image_layer = image_layer
        self.mask_layer = mask_layer

        self.bounding_box_layer = bounding_box_layer
        self.setObjectName("Objects")
        self.update_items()
        self.itemClicked.connect(self.select_item)
        self.installEventFilter(self)
        self.viewer.bind_key('d', overwrite=True)(self.toggle_bb_visibility)
        # bounding_box_layer.events.set_data.connect(on_data_change)

    @property
    def bounding_box_layer(self):
        return self._bounding_box_layer

    @bounding_box_layer.setter
    def bounding_box_layer(self, new_layer):
        if self._bounding_box_layer == new_layer:
            self.update_items()
            return
        self._bounding_box_layer = new_layer
        if new_layer is not None:
            if self.bounding_box_change not in self._bounding_box_layer.mouse_drag_callbacks:
                self._bounding_box_layer.mouse_drag_callbacks.append(self.bounding_box_change)
            if self._on_bb_double_click not in self._bounding_box_layer.mouse_double_click_callbacks:
                self._bounding_box_layer.mouse_double_click_callbacks.append(self._on_bb_double_click)
            # self._bounding_box_layer.events.connect(self.on_layer_event)
            if "label" not in self._bounding_box_layer.current_properties:
                self._bounding_box_layer.current_properties |= {"label": 0}
            self._bounding_box_layer.text = {
                "text": "{label:d}",
                "size": 10,
                "color": "green"
            }
        self.update_items()

    @property
    def image_layer(self):
        return self._image_layer

    @image_layer.setter
    def image_layer(self, new_layer):
        self._image_layer = new_layer
        for i in range(self.count()):
            item = self.item(i)
            item.image_layer = new_layer
            item.update_icon()
        self.update_items()

    @property
    def mask_layer(self):
        return self._mask_layer

    @mask_layer.setter
    def mask_layer(self, new_layer):
        self._mask_layer = new_layer
        for i in range(self.count()):
            item = self.item(i)
            item.mask_layer = new_layer
            item.update_icon()
        self.update_items()

    @property
    def indices(self):
        return self._indices

    @indices.setter
    def indices(self, indices):
        if indices is None:
            indices = 1
        if type(indices) is int:
            starting_value = indices
            def index():
                idx = starting_value
                while True:
                    yield idx
                    idx += 1
            indices = index()
        self._indices = itertools.cycle(iter(indices))

    def on_layer_event(self, event):
        if event.type == "set_data" and not self._mouse_down:
            self.update_items(True)

    def select_item(self, item):
        selected_idx = self.indexFromItem(item).row()
        if self.mask_layer is not None and self.image_layer is not None:
            item.on_select()
        else:
            self.clearSelection()
            self.clearFocus()
            warnings.warn("Image and labels layer should be selected!")
            return
        if selected_idx == self.selected_idx or self.mask_layer is None or self.image_layer is None:
            self.selected_idx = None
            item.unselect()
            self.clearSelection()
            self.clearFocus()
            self.bounding_box_layer.edge_color[selected_idx] = self.previous_edge_color
            self.bounding_box_layer.data = self.bounding_box_layer.data
        else:
            if self.selected_idx is not None:
                self.bounding_box_layer.edge_color[self.selected_idx] = self.previous_edge_color
            self.selected_idx = selected_idx
            self.previous_edge_color = self.bounding_box_layer.edge_color[selected_idx].copy()
            self.bounding_box_layer.edge_color[selected_idx] = (1., 0., 0., 1.)
            self.bounding_box_layer.data = self.bounding_box_layer.data

    def bounding_box_corners(self):
        if self._bounding_box_layer is None or len(self._bounding_box_layer.data) == 0:
            return []
        b_data = np.round(np.asarray(self._bounding_box_layer.data)).astype(int)
        if self.image_layer is not None:
            b_data -= self.image_layer.translate.astype(int)
        return np.concatenate([b_data.min(1, keepdims=True), b_data.max(1, keepdims=True)], axis=1)

    def eventFilter(self, source: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Close:
            if self.crop_image_layer is not None:
                self.viewer.layers.remove(self.crop_image_layer)
                self.viewer.layers.remove(self.crop_mask_layer)
                self.viewer.window.remove_dock_widget(self.projections_widget)
        elif event.type() == QEvent.ContextMenu and source is self:
            menu = QMenu()
            menu.addAction("Create layers")
            if menu.exec_(event.globalPos()):
                item = self.itemAt(event.pos())
                item.create_layers()
            return True
        elif event.type() == QEvent.Enter and type(source) == QObjectWidget:
            item = source.parent
            if item is None:
                return False
            idx = self.indexFromItem(item).row()
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self.bounding_box_layer._value = (idx, None)
                self.bounding_box_layer._set_highlight()
            return True
        elif event.type() == QEvent.Leave and type(source) == QObjectWidget:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self.bounding_box_layer._value = (None, None)
                self.bounding_box_layer._set_highlight()
            return True
        return super().eventFilter(source, event)

    def hideEvent(self, a0: QHideEvent) -> None:
        self.viewer.bind_key('d')(None)
        super().hideEvent(a0)

    def toggle_bb_visibility(self, _=None):
        if self.bounding_box_layer is not None and self.bounding_box_layer.visible:
            mode = self.bounding_box_layer.mode
            self.bounding_box_layer.visible = False
            yield
            self.bounding_box_layer.visible = True
            self.bounding_box_layer.mode = mode

    def bounding_box_change(self, layer, event):
        previous_data = np.asarray([bb.copy() for bb in self.bounding_box_layer.data])
        self._mouse_down = True
        yield
        if len(layer.data)>len(previous_data):
            self.bounding_box_layer.features["label"].iat[-1] = self.next_index()
            text = dict(self.bounding_box_layer.text)
            if napari.__version__ == "0.4.15":
                del text["values"]
            text["text"] = "{label:d}"
            self.bounding_box_layer.text = text
        while event.type == "mouse_move":
            yield
        self._mouse_down = False
        new_data = np.asarray(self.bounding_box_layer.data)
        im_size = self.image_layer.data.shape
        if self.image_layer.rgb:
            im_size = im_size[:-1]
        if len(new_data) > 0 and (np.any(np.all(new_data > im_size, 1))\
                or np.any(np.all(new_data < 0, 1))):
            self.bounding_box_layer.data = previous_data
        if np.shape(previous_data) != np.shape(self.bounding_box_layer.data):
            if len(new_data) > len(previous_data):
                self.bounding_box_layer.data = [np.clip(data, 0, np.asarray(im_size) - 1) for data in new_data]
        self.update_items()
        if self.projections_widget is not None:
            for p in self.projections_widget.projections:
                p.update_overlay()
                p.update()

    def update_items(self, update_all_icons=False):
        bounding_boxes = self.bounding_box_corners()
        if len(bounding_boxes) == 0:
            self.clear()
        elif len(bounding_boxes) < self.count():
            previous_data = np.asarray(list(np.unique(self.item(i).bounding_box, axis=0) for i in range(self.count())))
            removed_idx = np.argwhere(np.all(np.any(~np.equal(previous_data[np.newaxis], bounding_boxes[:, np.newaxis]), (-2, -1)), axis=0))
            removed_idx = np.squeeze(removed_idx)
            if removed_idx.ndim == 0:
                self.takeItem(removed_idx)
            else:
                self.clear()
        for i, bb in enumerate(bounding_boxes):
            curr_item = self.item(i)
            if curr_item is not None:
                if not np.allclose(curr_item.bounding_box, bb):
                    curr_item.bounding_box = bb
                if curr_item.image_layer is not self.image_layer:
                    curr_item.image_layer = self.image_layer
                if curr_item.idx != self.bounding_box_layer.features["label"][i]:
                    curr_item.idx = self.bounding_box_layer.features["label"][i]
                if update_all_icons:
                    curr_item.update_icon()

            else:
                new_item = QObjectListWidgetItem(self.name_template, bb, self.bounding_box_layer.features["label"][i], self, self.viewer, self.image_layer, self.mask_layer, self.channels_dim)
                self.insertItem(i, new_item)

    def insertItem(self, index, item: QObjectListWidgetItem):
        super().insertItem(index, item)
        self.setItemWidget(item, item.object_item)

    def addItem(self, item: QObjectListWidgetItem):
        self.insertItem(self.count(), item)

    def next_index(self):
        return next(self.indices)

    def _on_bb_double_click(self, layer=None, event=None):
        if len(self.bounding_box_layer.selected_data) == 0:
            return
        idx = next(iter(self.bounding_box_layer.selected_data))
        item = self.item(idx)
        self.setCurrentItem(item)
        self.select_item(item)


class ListWidgetBB(WidgetWithLayerList):
    def __init__(self, viewer: Viewer):
        super().__init__(viewer, [("bounding_box", BoundingBoxLayer), ("image", Image), ("labels", Labels)], False)
        layout = QVBoxLayout()
        self.viewer = viewer
        self.prev_n_layers = len(viewer.layers)
        self.channels_dim = None
        self.list_widget = None
        self.reset_index()
        self.bounding_box.combobox.currentIndexChanged.connect(self.bb_index_change)
        self.image.combobox.currentIndexChanged.connect(self.img_index_change)
        self.labels.combobox.currentIndexChanged.connect(self.mask_index_change)
        layout.addWidget(self.bounding_box.combobox)
        layout.addWidget(self.image.combobox)
        layout.addWidget(self.labels.combobox)

        self.list_widget_container = QStackedWidget()
        no_data_widget = QWidget()
        no_data_layout = QVBoxLayout()
        no_data_layout.addWidget(QLabel("No bounding box layer selected"))
        add_bb_button = QPushButton("Create Bounding Box layer")
        add_bb_button.clicked.connect(self.create_bb_layer)
        no_data_layout.addWidget(add_bb_button)
        no_data_widget.setLayout(no_data_layout)
        self.list_widget_container.addWidget(no_data_widget)
        self.list_widget_container.setCurrentIndex(0)
        layout.addWidget(self.list_widget_container)
        buttons_widget = QWidget()
        buttons_layout = QHBoxLayout()
        import_button = QPushButton("Import")
        import_button.clicked.connect(self.import_bounding_boxes)
        export_button = QPushButton("Export")
        export_button.clicked.connect(self.export_bounding_boxes)
        from_labels_button = QPushButton("From labels")
        from_labels_button.clicked.connect(self.bounding_boxes_from_labels)
        buttons_layout.addWidget(import_button)
        buttons_layout.addWidget(export_button)
        buttons_layout.addWidget(from_labels_button)
        buttons_widget.setLayout(buttons_layout)
        layout.addWidget(buttons_widget)
        self.setLayout(layout)
        self.name_template = "Object"
        if self.bounding_box.layer is not None:
            self.create_list_widget()
        self.installEventFilter(self)

    def bb_index_change(self, index):
        if self.bounding_box.layer is None:
            self.reset_index()
            self.remove_list_widget()
            return
        if self.list_widget is None:
            self.create_list_widget()
        self.list_widget.bounding_box_layer = self.bounding_box.layer
        if "label" in self.bounding_box.layer.features:
            self.reset_index(max(self.bounding_box.layer.features["label"]) + 1)

    def mask_index_change(self, index):
        if self.list_widget is not None and self.list_widget.mask_layer is not self.labels.layer:
            self.list_widget.mask_layer = self.labels.layer

    def img_index_change(self, index):
        self.update_channels_dim(index)

    @property
    def image_layer(self):
        return self.image.layer

    @property
    def mask_layer(self):
        return self.labels.layer

    def create_bb_layer(self, _):
        if self.image_layer is not None:
            self.viewer.add_layer(
                BoundingBoxLayer(ndim=self.image_layer.ndim, edge_color="green", face_color="transparent"))

    def update_channels_dim(self, idx):
        if idx == 0 or self.image_layer is None:
            return
        smallest_dim = np.argsort(self.image_layer.data.shape)[0]
        if self.image_layer.data.shape[smallest_dim] <= 3:
            self.channels_dim = smallest_dim
        else:
            self.channels_dim = None
        if self.list_widget is not None:
            self.list_widget.image_layer = self.image_layer
            self.list_widget.channels_dim = self.channels_dim

    def create_list_widget(self):
        self.list_widget = ObjectListWidget(self.viewer, self.name_template, self.bounding_box.layer, self.image_layer, self.labels.layer, self.channels_dim, 1)
        self.list_widget_container.addWidget(self.list_widget)
        self.list_widget_container.setCurrentIndex(1)

    def remove_list_widget(self):
        while self.list_widget_container.count() > 1:
            self.list_widget_container.removeWidget(self.list_widget_container.widget(1))
        self.list_widget_container.setCurrentIndex(0)
        self.list_widget = None

    def export_bounding_boxes(self):
        if self.bounding_box.layer is None:
            return
        filename = QFileDialog.getSaveFileName(self, "Select...", None, "(*.txt)")[0]
        bboxes = np.asarray(list(map(lambda bb: np.concatenate([bb.min(axis=0), bb.max(axis=0)]), self.bounding_box.layer.data)))
        with open(filename, "w") as f:
            writer = csv.writer(f)
            for i, bb in enumerate(bboxes):
                item = self.list_widget.item(i)
                writer.writerow([item.name, item.idx] + list(bb))

    def import_bounding_boxes(self):
        filename = QFileDialog.getOpenFileName(self, "Select...", None, "(*.txt)")[0]
        if not filename:
            return
        names = []
        idxs = []
        data = []
        with open(filename, "r") as f:
            reader = csv.reader(f)
            for line in reader:
                if len(line) == 0:
                    continue
                names.append(line[0])
                idxs.append(int(line[1]))
                data.append(list(map(lambda x: float(x), line[2:])))
        data = np.asarray(data)
        bounding_box_corners = np.reshape(data, (len(data), 2, -1))
        mask = np.asarray(list(itertools.product((False, True), repeat=bounding_box_corners[0].shape[1])))
        bounding_boxes = np.asarray([np.where(mask, bbc[1], bbc[0]) for bbc in bounding_box_corners])
        bounding_box_layer = BoundingBoxLayer(bounding_boxes, name=os.path.basename(filename), edge_color="green", face_color="transparent", features={"label": idxs})
        self.viewer.add_layer(bounding_box_layer)
        self.bounding_box.layer = bounding_box_layer

        for i, (name, idx) in enumerate(zip(names, idxs)):
            item = self.list_widget.item(i)
            item.name = name
            item.idx = idx
        self.reset_index(max(idxs) + 1)

    def bounding_boxes_from_labels(self):
        if self.labels.layer is None:
            return
        bb_layer = BoundingBoxLayer(ndim=self.image_layer.ndim, edge_color="green", face_color="transparent")
        bb_corners = find_objects(self.labels.layer.data)
        ids = []
        bbs = []
        for i, bb in enumerate(bb_corners):
            if bb is None:
                continue
            min_ = [slice_.start for slice_ in bb]
            max_ = [slice_.stop - 1 for slice_ in bb]
            bb = np.asarray(np.where(list(itertools.product((False, True), repeat=self.image_layer.ndim)), max_, min_))
            bbs.append(bb)
            ids.append(i+1)
        bb_layer.data = bbs
        bb_layer.features["label"] = ids
        self.viewer.add_layer(bb_layer)
        self.bounding_box.layer = bb_layer


    def reset_index(self, starting_index=1):
        if self.list_widget is not None:
            self.list_widget.indices = starting_index



import time

import napari
import numpy as np
import cv2
from napari import Viewer
from napari.layers import Labels
from scipy.interpolate import interp1d
from scipy.ndimage.measurements import find_objects
from qtpy.QtWidgets import QWidget, QVBoxLayout, QLabel, QSpinBox, QPushButton, QComboBox
from qtpy.QtCore import QThread, QObject, Signal

from ._utils.progress_widget import ProgressWidget
from ..mean_contour import settings
from ..mean_contour.meanContour import MeanThread
from ..mean_contour._contour import calcRpsvInterpolation
from ..mean_contour._reconstruction import reconstruct
from ..mean_contour._essentials import magnitude


def contour_cv2_mask_uniform(mask, contoursize_max):
    mask = mask.astype(np.uint8)
    contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    areas = [cv2.contourArea(cnt) for cnt in contours]
    max_ind = np.argmax(areas)
    contour = np.squeeze(contours[max_ind])
    contour = np.reshape(contour, (-1, 2))
    contour = np.append(contour, contour[0].reshape((-1, 2)), axis=0)
    contour = contour.astype('float32')

    rows, cols = mask.shape
    delta = np.diff(contour, axis=0)
    s = [0]
    for d in delta:
        dl = s[-1] + np.linalg.norm(d)
        s.append(dl)

    if (s[-1] == 0):
        s[-1] = 1

    s = np.array(s) / s[-1]
    fx = interp1d(s, contour[:, 0] / rows, kind='linear')
    fy = interp1d(s, contour[:, 1] / cols, kind='linear')
    S = np.linspace(0, 1, contoursize_max, endpoint=False)
    X = rows * fx(S)
    Y = cols * fy(S)

    contour = np.transpose(np.stack([X, Y])).astype(np.float32)

    contour = np.stack((contour[:, 1], contour[:, 0]), axis=-1)
    return contour


class InterpolationWorker(QObject):
    done = Signal("PyQt_PyObject")
    progress = Signal(int)
    dimension: int
    n_contour_points: int
    data: np.ndarray
    use_rpsv: bool
    max_iterations: int
    dims_displayed: tuple
    current_step: tuple
    selected_label: int

    def run(self):
        try:
            dimension = self.dimension
            n_contour_points = self.n_contour_points
            data = self.data.copy()
            use_rpsv = self.use_rpsv
            layer_slice_template = [
                slice(None) if d in self.dims_displayed
                    else None if d == dimension
                    else self.current_step[d]
                for d in range(data.ndim)]
            prev_cnt = None
            prev_layer = None
            start = time.time()
            for i in range(data.shape[dimension]):
                self.progress.emit(i)
                layer_slice = layer_slice_template.copy()
                layer_slice[dimension] = i
                mask = data[tuple(layer_slice)] == self.selected_label
                mask = mask.astype(np.uint8)
                if mask.max() == 0:
                    continue
                cnt = contour_cv2_mask_uniform(mask, n_contour_points)
                centroid = cnt.mean(0)
                start_index = np.argmin(np.abs(np.arctan2(*(cnt - centroid).T)))
                cnt = np.roll(cnt, -start_index, 0)
                if prev_cnt is not None:
                    if use_rpsv:
                        stgs = settings.Settings(max_iterations=self.max_iterations,
                                                 n_points=n_contour_points)
                        rpsv_thread = MeanThread([prev_cnt, cnt], stgs)
                        rpsv_thread.run()
                    for j in range(prev_layer + 1, i):
                        inter_layer_slice = layer_slice_template.copy()
                        inter_layer_slice[dimension] = j
                        prev_w = i - j
                        cur_w = j - prev_layer
                        weights = [prev_w, cur_w]
                        if use_rpsv:
                            contours = rpsv_thread.contours
                            regularMean = np.zeros_like(contours[0].lookup[contours[0].parameterization, :])
                            for j in range(2):
                                regularMean += contours[j].lookup[contours[j].parameterization, :] * weights[j]
                            regularMean /= np.sum(weights)
                            q_mean = calcRpsvInterpolation(contours, weights)
                            guessRayLengths = np.zeros(contours[0].lookup[contours[0].parameterization].shape[0])
                            for i_contour in range(2):
                                contourtmp = contours[i_contour].lookup[contours[i_contour].parameterization]
                                contourlengths = magnitude(contourtmp)
                                guessRayLengths += contourlengths * weights[i_contour]
                            guessRayLengths /= np.sum(weights)
                            guessRayLengths = magnitude(regularMean)

                            qraylengths = magnitude(q_mean)
                            qraylengths[qraylengths < 1e-99] = 1e-99

                            dirs = q_mean / qraylengths.reshape(qraylengths.shape[0], 1)
                            r_mean_lengths, costs = reconstruct(q_mean, guessRayLengths.copy(), stgs, rpsv_thread.rpSignal)
                            mean_cnt = dirs * r_mean_lengths.reshape(r_mean_lengths.shape[0], 1)
                        else:
                            mean_cnt = (prev_w * prev_cnt + cur_w * cnt)/(prev_w + cur_w)
                        mean_cnt = mean_cnt.astype(np.int32)
                        mask = np.zeros_like(data[tuple(inter_layer_slice)])
                        cv2.drawContours(mask, [np.flip(mean_cnt, -1)], 0, self.selected_label, -1)
                        cur_slice = data[tuple(inter_layer_slice)]
                        cur_slice[mask > 0] = mask[mask > 0]
                prev_cnt = cnt
                prev_layer = i
            napari.notification_manager.receive_info("Done in %.4f s" % (time.time() - start))
            self.done.emit(data)
        except Exception as e:
            self.done.emit(None)
            raise e


class InterpolationWidget(QWidget):
    def __init__(self, viewer: Viewer):
        super().__init__()
        self.viewer = viewer
        self.progress_dialog = ProgressWidget(message="Interpolating slices...")
        layout = QVBoxLayout()
        self.active_labels_layer = None

        self.interpolation_thread = QThread()
        self.interpolation_worker = InterpolationWorker()
        self.interpolation_worker.moveToThread(self.interpolation_thread)
        self.interpolation_worker.done.connect(self.interpolation_thread.quit)
        self.interpolation_worker.done.connect(lambda _: self.progress_dialog.setVisible(False))
        self.interpolation_worker.done.connect(self.set_labels)
        self.interpolation_worker.progress.connect(self.progress_dialog.setValue)
        self.interpolation_thread.started.connect(self.interpolation_worker.run)
        layout.addWidget(QLabel("dimension"))
        self.dimension_dropdown = QSpinBox()
        self.dimension_dropdown.setMinimum(0)
        self.dimension_dropdown.setMaximum(0)
        layout.addWidget(self.dimension_dropdown)

        layout.addWidget(QLabel("Method"))
        self.method_dropdown = QComboBox()
        self.method_dropdown.addItem("Geometric mean")
        self.method_dropdown.addItem("RPSV")
        self.method_dropdown.currentTextChanged.connect(lambda _: self.rpsv_widget.setVisible(self.method_dropdown.currentText() == "RPSV"))
        layout.addWidget(self.method_dropdown)

        layout.addWidget(QLabel("# contour points"))
        self.n_points = QSpinBox()
        self.n_points.setMinimum(10)
        self.n_points.setMaximum(1000)
        self.n_points.setValue(300)
        layout.addWidget(self.n_points)

        self.rpsv_widget = QWidget()
        rpsv_layout = QVBoxLayout()
        rpsv_layout.addWidget(QLabel("max iterations"))
        self.rpsv_iterations_spinbox = QSpinBox()
        self.rpsv_iterations_spinbox.setMaximum(100)
        self.rpsv_iterations_spinbox.setMinimum(1)
        self.rpsv_iterations_spinbox.setValue(20)
        rpsv_layout.addWidget(self.rpsv_iterations_spinbox)
        self.rpsv_widget.setLayout(rpsv_layout)
        self.rpsv_widget.setVisible(self.method_dropdown.currentText() == "RPSV")
        rpsv_layout.setContentsMargins(0, 0, 0, 0)
        self.rpsv_widget.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.rpsv_widget)


        self.interpolate_button = QPushButton("Interpolate")
        self.interpolate_button.clicked.connect(self.interpolate)
        self.interpolation_worker.done.connect(lambda: self.interpolate_button.setEnabled(True))
        layout.addWidget(self.interpolate_button)
        self.viewer.layers.selection.events.connect(self.on_layer_selection_change)
        viewer.dims.events.order.connect(self.on_order_change)
        viewer.dims.events.ndisplay.connect(lambda _: self.interpolate_button.setEnabled(viewer.dims.ndisplay == 2))
        self.setLayout(layout)
        layout.addStretch()
        self.on_layer_selection_change()

    def on_layer_selection_change(self, event=None):
        if event is None or event.type == "changed":
            active_layer = self.viewer.layers.selection.active
            if isinstance(active_layer, Labels):
                self.active_labels_layer = active_layer
                self.dimension_dropdown.setMaximum(self.active_labels_layer.ndim)

    def set_has_channels(self, has_channels):
        self.channels_dim_dropdown.setEnabled(has_channels)

    def on_order_change(self, event):
        extents = self.active_labels_layer.extent.data[1] + 1
        not_displayed = extents[list(self.viewer.dims.not_displayed)]
        ch_excluded = list(filter(lambda x: x > 3, not_displayed))
        if len(ch_excluded) == 0:
            new_dim = self.viewer.dims.not_displayed[0]
        else:
            new_dim = self.viewer.dims.order[int(np.argwhere(not_displayed == ch_excluded[0])[0])]
        self.dimension_dropdown.setValue(new_dim)

    def interpolate(self):
        if self.active_labels_layer is None:
            return
        dimension = self.dimension_dropdown.value()
        self.progress_dialog.setMaximum(self.active_labels_layer.data.shape[dimension])
        self.progress_dialog.setVisible(True)
        self.interpolate_button.setEnabled(False)
        self.interpolation_worker.dimension = dimension
        self.interpolation_worker.n_contour_points = self.n_points.value()
        self.interpolation_worker.data = self.active_labels_layer.data
        self.interpolation_worker.use_rpsv = self.method_dropdown.currentText() == "RPSV"
        self.interpolation_worker.dims_displayed = self.viewer.dims.displayed
        self.interpolation_worker.current_step = self.viewer.dims.current_step
        self.interpolation_worker.selected_label = self.active_labels_layer.selected_label
        self.interpolation_worker.max_iterations = self.rpsv_iterations_spinbox.value()
        self.interpolation_thread.start()

    def set_labels(self, data):
        if data is None:
            return
        update_mask = data > 0
        if self.active_labels_layer.preserve_labels:
            update_mask &= self.active_labels_layer.data == 0
        self.active_labels_layer.data[update_mask] = data[update_mask]
        self.active_labels_layer.refresh()
